using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapOpenedMailboxInfoOperationsWaitingTests {
    [Test]
    public void TestIdle_IdleCapable()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        ThreadPool.QueueUserWorkItem(delegate(object state) {
          server.WaitForClientRequest(); // IDLE

          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n");

          server.WaitForClientRequest(); // DONE

          Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");

          // DONE
          server.EnqueueTaggedResponse("$tag OK done\r\n");
        });

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        var timeout = TimeSpan.FromSeconds(0.5);

        Smdn.Net.TestUtils.Elapses(timeout, delegate {
          mailbox.Idle(timeout);
        });

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");
      });
    }

    [Test]
    [TestCase(false)]
    [TestCase(true)]
    public void TestIdle_IdleCapable_ProcessSizeAndStatusResponseOnlyOnce_Expunge(bool asPushResponse)
    {
      var selectResp =
        "* EXISTS 3\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.AreEqual(3, mailbox.ExistMessageCount);

        var eventExistMessageCountChangedRaisedCount = 0;

        mailbox.Client.ExistMessageCountChanged += delegate(object sender, ImapMailboxSizeChangedEventArgs e) {
          eventExistMessageCountChangedRaisedCount++;
        };

        ThreadPool.QueueUserWorkItem(delegate(object state) {
          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n");

          if (asPushResponse)
            server.EnqueuePushResponse("* EXPUNGE 2\r\n");

          server.WaitForClientRequest(); // DONE

          Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");

          // DONE
          if (asPushResponse)
            server.EnqueueTaggedResponse("$tag OK done\r\n");
          else
            server.EnqueueTaggedResponse("* EXPUNGE 2\r\n" +
                                         "$tag OK done\r\n");
        });

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        var timeout = TimeSpan.FromSeconds(0.5);

        Smdn.Net.TestUtils.Elapses(timeout, delegate {
          mailbox.Idle(timeout);
        });

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");

        Assert.AreEqual(2, mailbox.ExistMessageCount);
        Assert.AreEqual(1, eventExistMessageCountChangedRaisedCount);
      });
    }

    [Test]
    [TestCase(false)]
    [TestCase(true)]
    public void TestIdle_IdleCapable_ProcessSizeAndStatusResponseOnlyOnce_Recent(bool asPushResponse)
    {
      var selectResp =
        "* RECENT 0\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.AreEqual(0, mailbox.RecentMessageCount);

        var eventRecentMessageCountChangedRaisedCount = 0;

        mailbox.Client.RecentMessageCountChanged += delegate(object sender, ImapMailboxSizeChangedEventArgs e) {
          eventRecentMessageCountChangedRaisedCount++;
        };

        ThreadPool.QueueUserWorkItem(delegate(object state) {
          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n");

          if (asPushResponse)
            server.EnqueuePushResponse("* RECENT 1\r\n");

          server.WaitForClientRequest(); // DONE

          Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");

          // DONE
          if (asPushResponse)
            server.EnqueueTaggedResponse("$tag OK done\r\n");
          else
            server.EnqueueTaggedResponse("* RECENT 1\r\n" +
                                         "$tag OK done\r\n");
        });

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        var timeout = TimeSpan.FromSeconds(0.5);

        Smdn.Net.TestUtils.Elapses(timeout, delegate {
          mailbox.Idle(timeout);
        });

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");

        Assert.AreEqual(1, mailbox.RecentMessageCount);
        Assert.AreEqual(1, eventRecentMessageCountChangedRaisedCount);
      });
    }

    [Test]
    public void TestIdle_IdleCapable_TimeoutZero()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var delay = TimeSpan.FromSeconds(0.1);

        ThreadPool.QueueUserWorkItem(delegate(object state) {
          server.WaitForClientRequest(); // IDLE

          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n");

          Thread.Sleep(delay);

          // DONE
          server.EnqueueTaggedResponse("$tag OK done\r\n");
        });

        Smdn.Net.TestUtils.Elapses(delay, delegate {
          mailbox.Idle(0);
        });
      });
    }

    [Test]
    public void TestIdle_IdleCapable_DisconnectFromServer()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var delay = TimeSpan.FromSeconds(0.1);

        ThreadPool.QueueUserWorkItem(delegate(object state) {
          server.WaitForClientRequest(); // IDLE

          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n");

          Thread.Sleep(delay);

          server.EnqueueUntaggedByeResponse("* BYE Server shutting down.\r\n");
        });

        Smdn.Net.TestUtils.Elapses(delay, delegate {
          Assert.Throws<ImapConnectionException>(() => mailbox.Idle(TimeSpan.FromSeconds(1.0)));
        });
      });
    }

    [Test]
    public void TestIdle_IdleIncapable()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   IdleUnavailable);
    }

    [Test]
    public void TestIdle_IdleDisabled()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        mailbox.Client.DisabledCapabilities.Add(ImapCapability.Idle);

        IdleUnavailable(server, mailbox);
      });
    }

    private void IdleUnavailable(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox)
    {
      // NOOP (initial)
      server.EnqueueTaggedResponse("NOOP\r\n",
                                   "$tag OK done\r\n");
      // NOOP (first polling)
      server.EnqueueTaggedResponse("NOOP\r\n",
                                   "$tag OK done\r\n");
      // NOOP (stopped)
      server.EnqueueTaggedResponse("NOOP\r\n",
                                   "$tag OK done\r\n");

      var timeout = TimeSpan.FromSeconds(0.5);

      Smdn.Net.TestUtils.Elapses(timeout, delegate {
        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        mailbox.Idle(timeout, TimeSpan.FromSeconds(0.35));
      });

      Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");
    }

    [Test]
    public void TestIdle_IdleIncapable_TimeoutZero()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        // NOOP (initial)
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        // NOOP (stopped)
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");

        Smdn.Net.TestUtils.Elapses(TimeSpan.Zero, delegate {
          mailbox.Idle(0, 1000);
        });
      });
    }

    [Test]
    public void TestIdle_IdleIncapable_DisconnectFromServer()
    {
      TestUtils2.TestOpenedMailbox(new ImapCapability[0],
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        // NOOP (initial)
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        // NOOP (first polling)
        server.EnqueueUntaggedByeResponse("* BYE Server shutting down.\r\n");

        var timeout = TimeSpan.FromSeconds(2.5);
        var pollingInterval = TimeSpan.FromSeconds(0.25);

        Smdn.Net.TestUtils.Elapses(pollingInterval, delegate {
          Assert.Throws<ImapConnectionException>(() => mailbox.Idle(timeout, pollingInterval));
        });
      });
    }

    [Test]
    public void TestIdle_StopIdleWhileAwaiting()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        ThreadPool.QueueUserWorkItem(delegate(object state) {
          server.WaitForClientRequest();

          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n");

          Thread.Sleep(TimeSpan.FromSeconds(0.1));

          // DONE
          server.EnqueueTaggedResponse("$tag OK done\r\n");

          Assert.DoesNotThrow(() => mailbox.StopIdle());
        });

        var timeout = TimeSpan.FromSeconds(1.0);

        Smdn.Net.TestUtils.NotElapse(timeout, delegate {
          Assert.Throws<ImapProtocolViolationException>(() => mailbox.Idle(timeout));
        });

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestIdle_TimeoutInfinite()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.Idle(Timeout.Infinite));
      });
    }

    [Test]
    public void TestIdle_TimeoutArgumentOutOfRange()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.Idle(-2));
      });
    }

    [Test]
    public void TestStartIdle_IdleCapable_PollingIntervalInifinite()
    {
      StartIdleIdleCapablePollingIntervalIgnored(Timeout.Infinite);
    }

    [Test]
    public void TestStartIdle_IdleCapable_PollingIntervalArgumentOutOfRange_NegativeValue()
    {
      StartIdleIdleCapablePollingIntervalIgnored(-2);
    }

    [Test]
    public void TestStartIdle_IdleCapable_PollingIntervalArgumentOutOfRange_GreaterThan30Minutes()
    {
      StartIdleIdleCapablePollingIntervalIgnored((int)TimeSpan.FromMinutes(30.0).TotalMilliseconds);
    }

    private void StartIdleIdleCapablePollingIntervalIgnored(int millisecondsPollingInterval)
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        using (var waitHandle = new ManualResetEvent(false)) {
          ThreadPool.QueueUserWorkItem(delegate(object state) {
            server.WaitForClientRequest(); // IDLE

            server.EnqueueResponse("IDLE\r\n",
                                   "+ idling\r\n",
                                   () => waitHandle.Set(),
                                   null);

            server.WaitForClientRequest(); // DONE

            // DONE
            server.EnqueueTaggedResponse("$tag OK done\r\n");
          });

          var delay = TimeSpan.FromSeconds(0.25);

          Smdn.Net.TestUtils.Elapses(delay, delegate {
            Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

            mailbox.StartIdle(millisecondsPollingInterval);

            waitHandle.WaitOne();

            Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");

            Thread.Sleep(delay);

            mailbox.StopIdle();

            Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");
          });
        }
      });
    }

    [Test]
    public void TestStartIdle_IdleIncapable_PollingIntervalInfinite()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.StartIdle(Timeout.Infinite));
      });
    }

    [Test]
    public void TestStartIdle_IdleIncapable_PollingIntervalArgumentOutOfRange_NegativeValue()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.StartIdle(-2));
      });
    }

    [Test]
    public void TestStartIdle_IdleIncapable_PollingIntervalArgumentOutOfRange_GreaterThan30Minutes()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.StartIdle(TimeSpan.FromMinutes(30.0)));
      });
    }

    [Test]
    public void TestStartIdle_AnotherWatchOperationProceeding_1()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        using (var waitHandle = new ManualResetEvent(false)) {
          // first
          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n",
                                 () => waitHandle.Set(),
                                 null);

          mailbox.StartIdle();

          waitHandle.WaitOne();

          Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");
        }

        // second
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.StartIdle());

        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.StopIdle();

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");
      });
    }

    [Test]
    public void TestStartIdle_AnotherWatchOperationProceeding_2()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        using (var waitHandle = new ManualResetEvent(false)) {
          // first
          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n",
                                 () => waitHandle.Set(),
                                 null);

          mailbox.StartWaitForMessageArrival();

          waitHandle.WaitOne();

          Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");
        }

        // second
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.StartIdle());

        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.StopWaitForMessageArrival();

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");
      });
    }

    [Test]
    public void TestStopIdle_WatchOperationNotStarted()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.StopIdle());
      });
    }

    [Test]
    public void TestStopIdle_AnotherWatchOperationProceeding()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        using (var waitHandle = new ManualResetEvent(false)) {
          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n",
                                 () => waitHandle.Set(),
                                 null);

          mailbox.StartWaitForMessageArrival();

          waitHandle.WaitOne();

          Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");
        }

        Assert.Throws<ImapProtocolViolationException>(() => mailbox.StopIdle());

        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.StopWaitForMessageArrival();

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");
      });
    }

    [Test]
    public void TestWaitForMessageArrival_IdleCapable()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        ThreadPool.QueueUserWorkItem(delegate(object state) {
          server.WaitForClientRequest(); // IDLE

          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n");

          server.WaitForClientRequest(); // DONE

          // DONE
          server.EnqueueTaggedResponse("$tag OK done\r\n");
        });

        var timeout = TimeSpan.FromSeconds(0.5);
        ImapMessageInfoList arrivalMessages = null;

        Smdn.Net.TestUtils.Elapses(timeout, delegate {
          arrivalMessages = mailbox.WaitForMessageArrival(timeout);
        });

        Assert.IsNotNull(arrivalMessages);
        CollectionAssert.IsEmpty(arrivalMessages);
      });
    }

    [Test]
    public void TestWaitForMessageArrival_IdleCapable_TimeoutZero()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var delay = TimeSpan.FromSeconds(0.1);

        ThreadPool.QueueUserWorkItem(delegate(object state) {
          server.WaitForClientRequest(); // IDLE

          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n");

          Thread.Sleep(delay);

          // DONE
          server.EnqueueTaggedResponse("$tag OK done\r\n");
        });

        ImapMessageInfoList arrivalMessages = null;

        Smdn.Net.TestUtils.Elapses(delay, delegate {
          arrivalMessages = mailbox.WaitForMessageArrival(0);
        });

        Assert.IsNotNull(arrivalMessages);
        CollectionAssert.IsEmpty(arrivalMessages);
      });
    }

    [Test]
    public void TestWaitForMessageArrival_IdleCapable_TimeoutInfinite()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var delay = TimeSpan.FromSeconds(0.1);

        ThreadPool.QueueUserWorkItem(delegate(object state) {
          server.WaitForClientRequest(); // IDLE

          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n");

          Thread.Sleep(delay);

          server.EnqueuePushResponse("* RECENT 1\r\n");
          server.EnqueuePushResponse("* EXISTS 1\r\n");

          server.WaitForClientRequest();

          // DONE
          server.EnqueueTaggedResponse("$tag OK done\r\n");
        });

        ImapMessageInfoList arrivalMessages = null;

        Smdn.Net.TestUtils.Elapses(delay, delegate {
          arrivalMessages = mailbox.WaitForMessageArrival(Timeout.Infinite);
        });

        Assert.IsNotNull(arrivalMessages);

        server.EnqueueTaggedResponse("FETCH ",
                                     "* FETCH 1 (UID 1)\r\n" +
                                     "$tag OK done\r\n");

        var arrivalMessageArray = arrivalMessages.ToArray();

        Assert.AreEqual(1, arrivalMessageArray.Length);
        Assert.AreEqual(1L, arrivalMessageArray[0].Uid, "arrival message uid");
        Assert.AreEqual(1L, arrivalMessageArray[0].Sequence, "arrival message seq");
      });
    }

    [Test]
    public void TestWaitForMessageArrival_IdleIncapable()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        ThreadPool.QueueUserWorkItem(delegate(object state) {
          server.WaitForClientRequest();

          // NOOP (initial)
          server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");

          server.WaitForClientRequest();

          // NOOP (first polling)
          server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");

          server.WaitForClientRequest();

          // NOOP (stopped)
          server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        });

        var timeout = TimeSpan.FromSeconds(0.5);
        ImapMessageInfoList arrivalMessages = null;

        Smdn.Net.TestUtils.Elapses(timeout, delegate {
          arrivalMessages = mailbox.WaitForMessageArrival(timeout,
                                                          TimeSpan.FromSeconds(0.35));
        });

        Assert.IsNotNull(arrivalMessages);
        CollectionAssert.IsEmpty(arrivalMessages);
      });
    }

    [Test]
    public void TestWaitForMessageArrival_IdleIncapable_TimeoutZero()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        // NOOP (initial)
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        // NOOP (first polling)
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");

        var pollingInterval = TimeSpan.FromSeconds(0.25);
        ImapMessageInfoList arrivalMessages = null;

        Smdn.Net.TestUtils.NotElapse(pollingInterval, delegate {
          arrivalMessages = mailbox.WaitForMessageArrival(0,
                                                          (int)pollingInterval.TotalMilliseconds);
        });

        Assert.IsNotNull(arrivalMessages);

        var arrivalMessageArray = arrivalMessages.ToArray();

        Assert.AreEqual(0, arrivalMessageArray.Length);
      });
    }

    [Test]
    public void TestWaitForMessageArrival_IdleIncapable_TimeoutInfinite()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        // NOOP (initial)
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        // NOOP (first polling)
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        // NOOP (second polling)
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* RECENT 1\r\n" +
                                     "* EXISTS 1\r\n" +
                                     "$tag OK done\r\n");

        var pollingInterval = TimeSpan.FromSeconds(0.25);
        ImapMessageInfoList arrivalMessages = null;

        Smdn.Net.TestUtils.Elapses(TimeSpan.FromTicks(pollingInterval.Ticks * 2), delegate {
          arrivalMessages = mailbox.WaitForMessageArrival(Timeout.Infinite,
                                                          (int)pollingInterval.TotalMilliseconds);
        });

        Assert.IsNotNull(arrivalMessages);

        server.EnqueueTaggedResponse("FETCH ",
                                     "* FETCH 1 (UID 1)\r\n" +
                                     "$tag OK done\r\n");

        var arrivalMessageArray = arrivalMessages.ToArray();

        Assert.AreEqual(1, arrivalMessageArray.Length);
        Assert.AreEqual(1L, arrivalMessageArray[0].Uid, "arrival message uid");
        Assert.AreEqual(1L, arrivalMessageArray[0].Sequence, "arrival message seq");
      });
    }

    [Test]
    public void TestWaitForMessageArrival_TimeoutArgumentOutOfRange()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.WaitForMessageArrival(-2));
      });
    }

    [Test]
    public void TestWaitForMessageArrival_StopWaitForMessageArrivalWhileAwaiting()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 1]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        ThreadPool.QueueUserWorkItem(delegate(object state) {
          server.WaitForClientRequest(); // IDLE

          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n");

          Thread.Sleep(TimeSpan.FromSeconds(0.1));

          // DONE
          server.EnqueueTaggedResponse("$tag OK done\r\n");

          Assert.DoesNotThrow(() => mailbox.StopWaitForMessageArrival());
        });

        var timeout = TimeSpan.FromSeconds(1.0);

        Smdn.Net.TestUtils.NotElapse(timeout, delegate {
          Assert.Throws<ImapProtocolViolationException>(() => mailbox.WaitForMessageArrival(timeout));
        });
      });
    }

    [Test]
    public void TestStartWaitForMessageArrival_IdleCapable_PollingIntervalInifinite()
    {
      StartWaitForMessageArrivalIdleCapablePollingIntervalIgnored(-1);
    }

    [Test]
    public void TestStartWaitForMessageArrival_IdleCapable_PollingIntervalArgumentOutOfRange_NegativeValue()
    {
      StartWaitForMessageArrivalIdleCapablePollingIntervalIgnored(-2);
    }

    [Test]
    public void TestStartWaitForMessageArrival_IdleCapable_PollingIntervalArgumentOutOfRange_GreaterThan30Minutes()
    {
      StartWaitForMessageArrivalIdleCapablePollingIntervalIgnored((int)TimeSpan.FromMinutes(30.0).TotalMilliseconds);
    }

    private void StartWaitForMessageArrivalIdleCapablePollingIntervalIgnored(int millisecondsPollingInterval)
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        using (var waitHandle = new ManualResetEvent(false)) {
          ThreadPool.QueueUserWorkItem(delegate(object state) {
            server.WaitForClientRequest(); // IDLE

            server.EnqueueResponse("IDLE\r\n",
                                   "+ idling\r\n",
                                   () => waitHandle.Set(),
                                   null);

            server.WaitForClientRequest(); // DONE

            // DONE
            server.EnqueueTaggedResponse("$tag OK done\r\n");
          });

          var delay = TimeSpan.FromSeconds(0.25);

          Smdn.Net.TestUtils.Elapses(delay, delegate {
            mailbox.StartWaitForMessageArrival(millisecondsPollingInterval);

            waitHandle.WaitOne();

            Thread.Sleep(delay);

            mailbox.StopWaitForMessageArrival();
          });
        }
      });
    }

    [Test]
    public void TestStartWaitForMessageArrival_IdleIncapable_PollingIntervalInfinite()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.StartWaitForMessageArrival(Timeout.Infinite));
      });
    }

    [Test]
    public void TestStartWaitForMessageArrival_IdleIncapable_PollingIntervalArgumentOutOfRange_NegativeValue()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.StartWaitForMessageArrival(-2));
      });
    }

    [Test]
    public void TestStartWaitForMessageArrival_IdleIncapable_PollingIntervalArgumentOutOfRange_GreaterThan30Minutes()
    {
      TestUtils2.TestOpenedMailbox("INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.StartWaitForMessageArrival(TimeSpan.FromMinutes(30.0)));
      });
    }

    [Test]
    public void TestStartWaitForMessageArrival_AnotherWatchOperationProceeding1()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        using (var waitHandle = new ManualResetEvent(false)) {
          // first
          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n",
                                 () => waitHandle.Set(),
                                 null);

          mailbox.StartWaitForMessageArrival();

          waitHandle.WaitOne();

          Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");
        }

        // second
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.StartWaitForMessageArrival());

        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.StopWaitForMessageArrival();

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");
      });
    }

    [Test]
    public void TestStartWaitForMessageArrival_AnotherWatchOperationProceeding2()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        using (var waitHandle = new ManualResetEvent(false)) {
          // first
          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n",
                                 () => waitHandle.Set(),
                                 null);

          mailbox.StartIdle();

          waitHandle.WaitOne();

          Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");
        }

        // second
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.StartWaitForMessageArrival());

        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.StopIdle();

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");
      });
    }

    [Test]
    public void TestStopWaitForMessageArrival_WatchOperationNotStarted()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.StopWaitForMessageArrival());
      });
    }

    [Test]
    public void TestStopWaitForMessageArrival_AnotherWatchOperationProceeding()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        using (var waitHandle = new ManualResetEvent(false)) {
          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n",
                                 () => waitHandle.Set(),
                                 null);

          mailbox.StartIdle();

          waitHandle.WaitOne();

          Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");
        }

        Assert.Throws<ImapProtocolViolationException>(() => mailbox.StopWaitForMessageArrival());

        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.StopIdle();

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idling");
      });
    }

    [Test]
    public void TestWatchOperationProceeding_Refresh()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        using (var waitHandle = new ManualResetEvent(false)) {
          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n",
                                 () => waitHandle.Set(),
                                 null);

          mailbox.StartIdle();

          waitHandle.WaitOne();

          Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");
        }

        Assert.Throws<InvalidOperationException>(() => mailbox.Refresh());

        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.StopIdle();

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");
      });
    }

    [Test]
    public void TestWatchOperationProceeding_Close()
    {
      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Idle},
                                   "INBOX",
                                   "$tag OK [READ-WRITE] done\r\n",
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy before idle");

        using (var waitHandle = new ManualResetEvent(false)) {
          server.EnqueueResponse("IDLE\r\n",
                                 "+ idling\r\n",
                                 () => waitHandle.Set(),
                                 null);

          mailbox.StartIdle();

          waitHandle.WaitOne();

          Assert.IsTrue(mailbox.Client.IsBusy, "IsBusy while idling");
        }

        Assert.Throws<InvalidOperationException>(() => mailbox.Close());

        server.EnqueueTaggedResponse("$tag OK done\r\n");

        mailbox.StopIdle();

        Assert.IsFalse(mailbox.Client.IsBusy, "IsBusy after idle");
      });
    }
  }
}

